static char *RCSid="$Id: lafe.c,v 0.10 1995/11/25 06:21:28 dld Exp $";

/* Latency Free Empire

   This is a client optimized for high latency links.  You can type
   in, edit, and send several commands before the server replies with
   output from the first command.  This allows you to make better use
   of the existing bandwidth than most other clients.  pei and eif are
   particularly susceptible to latency, as they require a complete
   exchange of data before prompting again.  This could also be called
   LAg Free Empire.  Empire 2 asynch features are supported.  Empire 3
   C_SYNC isn't supported yet, as this is a rather dumb client (in
   between emp_client and eif).  Readline is supported, and at the
   moment required.  If you want lafe without readline, use emp_client
   instead.

   -harmless@empire.net */



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <sys/wait.h>
#include <signal.h>
#include <netdb.h>
#include <curses.h>
#include <term.h>
#include <fcntl.h>

#include <readline/readline.h>
#include <readline/history.h>


/* some defaults in case nothing is specified.  Change as you wish. */
#define DEFPORT "7778"
#define DEFHOST "empire.net"
#define DEFCOUNTRY "visitor"
#define DEFPASSWORD "visitor"



/************************************************************************
 *   The filedescriptor of the empire server socket.  Had to make this
 * global in order to pass to be accessible to event_hook().
 *   I always feel guilty when using globals...
 ************************************************************************/
static int s;


/* the length of the prompt.  This plus rl_point is how far we have to
   go back to the beginning of the line. */
extern int rl_visible_prompt_length;

/*  I'm cheating by using these.  */
extern int _rl_horizontal_scroll_mode;
extern int _rl_vis_botlin, _rl_last_v_pos, _rl_last_c_pos;

/* pointer to the readline prompt.  This can change in the middle of
 * readline() if a prompt arrives in the middle of typing. */
extern char *rl_prompt,*rl_display_prompt;


/***************************************************************************
 *   A few routines to clear the readline-portion of the display.
 *   Some initialization to get the necesary termcap strings and a
 *   couple support routines.
 ***************************************************************************/

static char *tc_sc, *tc_rc, *tc_cd; /* save, restore, clear-to-eos */
static char *tc_LE, *tc_ce;
static char *tc_ch, *tc_UP;

extern char *tgetstr();


void init_tc() {
  static char *termstrbuf;
  char **area;

  /* SGI and IBM docs say area is meaningless under terminfo emulation, but
     it crashes without a sizeable buffer on the SGI... */
  termstrbuf = malloc(2048);
  area = &termstrbuf;
  tgetent(termstrbuf,getenv("TERM"));

  tc_rc = tgetstr("rc",area); /* restore cursor */
  tc_cd = tgetstr("cd",area); /* clear eos */
  tc_sc = tgetstr("sc",area); /* save cursor */
  tc_LE = tgetstr("LE",area); /* left #1 */
  tc_ce = tgetstr("ce",area); /* clear to eol */
  tc_ch = tgetstr("ch",area); /* set horzontal cursor position */
  tc_UP = tgetstr("UP",area); /* up #1 */

  /* this section is just for debugging termcap capabilities */
#if 0
  if (tc_sc) fprintf(stderr,"tc: save cursor\n");
  else fprintf(stderr,"tc: no save cursor\n");

  if (tc_sc) fprintf(stderr,"tc: set horizontal cursor\n");
  else fprintf(stderr,"tc: no set horizontal cursor\n");

  if (tc_rc) fprintf(stderr,"tc: restore\n");
  else fprintf(stderr,"tc: no restore\n");

  if (tc_cd) fprintf(stderr,"tc: clear to end of screen\n");
  else fprintf(stderr,"tc: no clear to end of screen\n");

  if (tc_LE) fprintf(stderr,"tc: left param\n");
  else fprintf(stderr,"tc: no left param\n");

  if (tc_ce) fprintf(stderr,"tc: clear to end of line\n");
  else fprintf(stderr,"tc: no clear to end of line\n");
#endif

}

/* move to left edge and clear to end */
void clear_readline() {

  if (tc_ch) {
    tputs(tparm(tc_ch,0),1,putchar);
  } else if (tc_LE) {
    tputs(tparm(tc_LE,_rl_last_c_pos),1,putchar);
  } else if (tc_rc) {
    tputs(tc_rc,1,putchar);
  } else {
    putchar('\r');
  }

  if (_rl_last_v_pos>0) {
    if (tc_UP) {
      tputs(tparm(tc_UP,_rl_last_v_pos),1,putchar);
    } else {
      /* couldn't reach beginning */
    }
  }

  if (tc_cd) {
    tputs(tc_cd,1,putchar);
  } else if (tc_ce) {
    tputs(tc_ce,1,putchar);
    /* may not have cleared whole screen */
  } else {
    printf("\n");
  }

}

/* save position in case we need to clear */
void save_readline() {
  if (tc_sc) {
    tputs(tc_sc,1,putchar);
  }
}


void quit() {
  exit(0);
}

void execute(char *line) {
  int fd,len,wlen,offset=0,length=0;
  char buffer[4096];
  char *filename;

  for (filename=line;isspace(*filename);filename++) ;
  for (;!isspace(*filename);filename++) ;
  for (;isspace(*filename);filename++) ;

  fd=open(filename,O_RDONLY);
  
  if (fd<0) {
    sprintf(buffer,"open failed on '%s'",filename);
    perror(buffer);
    return;
  }
  while ( (len=read(fd,buffer,sizeof(buffer))) > 0 ) {
    length += len;

    write(fileno(stdout),buffer,len);

    do {
      wlen=write(s,buffer+offset,len-offset);
      if (wlen>0) {
	len -= wlen;
	offset+=wlen;
      } else {
	perror("exec write");
      }
    } while (len>0);
  }

  if (len<0) {
    perror("exec read");
  }

  close(fd);
}

  

/********************************************************
 *	Maximum number of commands on the output stack
 ********************************************************/
#define COMMAND_STACK 20
static int commands=0;
static char *commandline[COMMAND_STACK];
static FILE *out=stdout;
static char *prompt=NULL;
static int timeleft,btus,teles=0;


/************************************************************
 *	Handle input lines depending on protocol number
 ************************************************************/
void process_line(char *buffer,int len) {
  int i,j;

  if (len<3) return;

  switch (buffer[0]) {

  case '3':
  case 'a':
  case 'b':
    fwrite(buffer+2,1,len-2,out);
    quit();
    break;

  case '1':
  case 'd':
    fwrite(buffer+2,1,len-2,out);
    break;

  case '0':
  case '2':
  case '7':
    /* fwrite(buffer,1,len,out); */
    break;

  case 'e':
    buffer[len-1]=0;
    /* rl_message(buffer+2); */
    if (len>3) {
      teles=1;
      sscanf(buffer+2,"%d new teles",&teles);
      putchar('\07');

      if (prompt) free(prompt);
      prompt=malloc(len+30);
      sprintf(prompt,"%s [%d:%d] Command: ",buffer+2,timeleft,btus);

      /* cheated a bit here again.. this isn't supposed to be allowed */
      rl_display_prompt=rl_prompt=prompt;
      rl_visible_prompt_length = rl_expand_prompt(rl_prompt);

    } else {
      teles=0;
    }


    break;

  case '4':
  case '5':
    if (commands>0) commands--;
    if (commands>0) {
	buffer[len-2]=0;
	fprintf(stdout,"%s %s\n",buffer+2,commandline[commands]);
    }
    if (commandline[commands]) {
      free(commandline[commands]);
    }
    if (prompt) free(prompt);
    prompt=malloc(len);
    memcpy(prompt,buffer+2,len-3);
    prompt[len-3]=' ';
    prompt[len-2]=0;
    rl_display_prompt=rl_prompt=prompt;
    rl_visible_prompt_length = rl_expand_prompt(rl_prompt);
    break;

  case '6':
    if (out!=stdout) {
      fclose(out);
      /* if pipe */
      wait(NULL);
    }
    out=stdout;
    sscanf(buffer+2,"%d %d",&timeleft,&btus);
    if (commands>0) commands--;
    if (commands>0) {
      buffer[len-2]=0;
      fprintf(stdout,"[%d:%d] %s\n",timeleft,btus,commandline[commands]);
    }
    if (commandline[commands]) {
      free(commandline[commands]);
    }

    if (prompt) free(prompt);
    prompt=malloc(len+30);
    sprintf(prompt,"[%d:%d] Command: ",timeleft,btus);

    /* cheated a bit here again.. this isn't supposed to be allowed */
    rl_display_prompt=rl_prompt=prompt;
    rl_visible_prompt_length = rl_expand_prompt(rl_prompt);

    break;

  case '8':
    for (i=2;(buffer[i]=='>' || isspace(buffer[i])) && i<len;i++) ;
    for (j=i;!isspace(buffer[j]) && j<len;j++) ;
    buffer[j]=0;
    out = fopen(buffer+i,"w");
    break;

  case '9':
    for (i=2;(buffer[i]=='|' || isspace(buffer[i])) && i<len;i++) ;
    for (j=i;!isspace(buffer[j]) && j<len;j++) ;
    buffer[j]=0;
    out = popen(buffer+i,"w");
    break;

  case 'c':
    fprintf(stderr,"C_EXEC should have been processed locally\n");
    write(s,"ctld\n",5);

    break;

    

  case 'f':
    /* C_SYNC protocol ... ignored for now: toggle sync off */
    /* glue to the client-lib when released */
    /* fwrite(buffer,1,len,out); */
    break;

  default:
    fprintf(stderr,"Unimplemented protocol:\n");
    fwrite(buffer,1,len,out);
    break;
  }
}


/* process whole lines, transfer remainder to the top of the buffer
   and return the length of the remainder */
int output(char *buffer,int len) {
  int start,end;

  start=0;
  
  for (;;) {
    for (end=start;end<len;end++) {
      if (buffer[end]=='\n') break;
    }
    if (end<len) {
      process_line(buffer+start,end-start+1);
    } else {
      if (end!=start) {
	memcpy(buffer,buffer+start,end-start);
	buffer[end-start]=0;
	return end-start;
      } else {
	return 0;
      }
    }
    start=end+1;
  }
}

/***************************************************************************
 *     The guts of the asynchronous behavior of lbe.  Selects both the
 *  socket to the empire server and the tty to wait for input.
 *  Prints out any socket data (redrawing partial input lines if
 *  necessary).  Returns when tty input is found (so readline() can
 *  continue).
 *     The name of this routine is passed to readline through a global
 *  variable (rl_event_hook).
 ***************************************************************************/
void event_hook() {
  fd_set in;
  int i;
  int len;
  static char buffer[2048];
  static int offset=0;	/* unprocessed remnants in buffer */

  /* wait for something from stdin or s */
  FD_ZERO(&in);
  FD_SET(s,&in);
  FD_SET(fileno(stdin),&in);
  i=select( FD_SETSIZE,&in,NULL,NULL,NULL);

  if (i<0) {
    perror("select");
  }

  if (i>0) {

    if (FD_ISSET(s,&in)) {
      int received=0;

      clear_readline();
      fflush(rl_outstream);

      while ( (len=recv(s,buffer+offset,sizeof(buffer)-offset,0)) > 0) {
	received+=len;
	offset=output(buffer,len+offset);
      }

      if (received==0) {
	if (errno==ENOENT) printf("Lost connection to server\n");
	else perror("recv");
	quit();
      }

      if (errno==EAGAIN) {
	/* no worries */
      } else {
	perror("recv2");
	quit();
      }

      rl_forced_update_display();
    }
  }
}




/************************************************************
 *    Open a socket to host/port and return filedescriptor.
 ************************************************************/
int empconnect(char *host,int port,char *user,char *country,char *pass) {
  struct	hostent *hp;
  struct	sockaddr_in sin;
  char buffer[1024];

  if (isdigit(*host)) {
    sin.sin_addr.s_addr = inet_addr(host);
  } else {
    hp = gethostbyname(host);
    if (hp == NULL) {
      fprintf(stderr, "%s: No such host\n", host);
      return 0;
    }
    memcpy( &sin.sin_addr, hp->h_addr, sizeof(sin.sin_addr));
  }

  sin.sin_port = htons(port);

  s = socket(AF_INET, SOCK_STREAM, 0);
  if (s < 0) {
    perror("socket:");
    return -1;
  }
  sin.sin_family = AF_INET;
  if (connect(s, &sin, sizeof(sin)) < 0) {
    perror("connect");
    return -1;
  }

  fcntl(s,F_SETFL,FNONBLOCK);


  sprintf(buffer,"user %s\ncoun %s\npass %s\nplay\n",user,country,pass);
  write(s,buffer,strlen(buffer));

  return s;
}



/**************************************************************************
 *   parse commands.  Look for commands matching local commands
 *   and call the local handler.  All others are sent directly to
 *   the server on socket s.
 **************************************************************************/
struct local_command {
  char *name;	      /* command name .. "execute" */
  int shortest;       /* shortest abbreviation */
  void (*process)();  /* function to parse arguments */
};


static struct local_command local_command[] = {
  { "execute", 4, execute }
};

void parse_command(char *line) {
  int len; /* length of line */
  int loc; /* location in local_command[] */
  int pos; /* position in command string */
  int start;  /* start of command string */
  
  for (start=0;line[start] && isspace(line[start]);start++) ;

  loc=pos=0;

  while (loc<sizeof(local_command)/sizeof(*local_command)) {
    if (line[start+pos]==local_command[loc].name[pos]) {
      pos++;

      /* found a match, call handler and return */
      if (pos>=local_command[loc].shortest) {
	local_command[loc].process(line);
	return;
      }

    } else if (line[start+pos] > local_command[loc].name[pos]) {
      loc++;
      pos=0;
    } else {
      break;
    }
  }

  /* no match found, send to server */
  len=strlen(line);
  line[len]='\n';
  if (write(s,line,len+1) != len+1) {
    perror("command write failure");
  }
  line[len]=0;

  commandline[commands++]=line;
}


/***********************************************************************
 *    Set up readline() and socket s to server.  Push each new line
 *  down the socket.
 ***********************************************************************/
int main(int argc, char **argv)
{
  char *line;
  char *host,*port,*country,*password;

  rl_readline_name = argv[0];
  rl_event_hook = (Function *)event_hook;
  rl_startup_hook = (Function *)save_readline;


  init_tc();
  
  signal(SIGPIPE,SIG_IGN);	/* don't worry if a destination pipe closes */

  if (argc==5) {
    host=argv[1];
    port=argv[2];
    country=argv[3];
    password=argv[4];
  } else if (argc==3) {
    country=argv[1];
    password=argv[2];
    host=getenv("EMPIREHOST");
    port=getenv("EMPIREPORT");
  } else if (argc==1) {
    country=getenv("COUNTRY");
    password=getenv("REPRESENTATIVE");
    host=getenv("EMPIREHOST");
    port=getenv("EMPIREPORT");
  } else {
    fprintf(stderr,"Usage:\
\t%s host port country password\n\
\t%s host port\n\
\t%s\n\
environment variables: EMPIREHOST EMPIREPORT COUNTRY REPRESENTATIVE\n",
	    argv[0],argv[0],argv[0]);
  }

  if (!host) { 
    host=DEFHOST;
    fprintf(stderr,"host defaulting to %s\n",host);
  }
  if (!port) {
    port=DEFPORT;
    fprintf(stderr,"port defaulting to %s\n",port);
  }
  if (!country) {
    country=DEFCOUNTRY;
    fprintf(stderr,"country defaulting to %s\n",country);
  }
  if (!password) {
    password=DEFPASSWORD;
    fprintf(stderr,"password defaulting to %s\n",password);
  }

  s=empconnect(host,atoi(port),getenv("USER"),country,password);

  if (s<0) {
    fprintf(stderr,"Unable to log in to %s at %s/%s\n",country,host,port);
    exit(-1);
  }

  /* clear command arguments */
  for (;argc>0;argc--) argv[argc-1]=0;

  commands=1;
  prompt=NULL;
  
  /* Loop reading and executing lines until the user quits. */
  for ( ; ; ) {
    int len;

    if (!commands) {
      line = readline(prompt);
    } else {
      line = readline(": ");
    }

    if (!line)
      break;
    
    add_history(line);


    /* command handler, most get written directly to socket s */
    parse_command(line);

  }


  close(s);
  return 0;
}

